@@ -8,13 +8,10 @@ from typing import List, Tuple, Any, Dict, Callable, Union, Optional
 
 import xlrd
 from bson import json_util
-from pyspark.sql import SparkSession
-from pyspark.sql.functions import expr
-from pyspark.sql.types import *
 from xlrd import open_workbook, Book
 from xlrd.sheet import Sheet, Cell
 
-__all__ = ['SqlTester', 'TableData', 'TestCase', 'work_path', 'SqlReader', 'TableColumnTypes']
+__all__ = ['SqlTester', 'TableData', 'TestCase', 'WorkPath', 'work_path', 'SqlReader', 'TableColumnTypes']
 
 from easy_sql.sql_processor import SqlProcessor
 from easy_sql.sql_processor.backend import Backend, SparkBackend, Row, Partition
@@ -70,15 +67,15 @@ class lazy_property(object):
 
 class WorkPath:
 
-    def __init__(self):
-        self._work_path: Optional[str] = None
+    def __init__(self, work_path: str = None):
+        self._work_path: Optional[str] = work_path
 
     def set_work_path(self, work_path: str):
         self._work_path = work_path
 
     def work_path(self):
         if not self._work_path:
-            raise Exception('work_path must be set before use.')
+            raise AssertionError('work_path must be set before use.')
         return self._work_path
 
     def path(self, relative_path: str):
@@ -86,7 +83,7 @@ class WorkPath:
 
     def relative_path(self, abs_path: str):
         if not abs_path.startswith(self.work_path()):
-            raise Exception(f'abs path `{abs_path}` not in work directory: ')
+            raise AssertionError(f'abs path `{abs_path}` not in work directory: ')
         return '' if len(abs_path) == self.work_path() else abs_path[len(self.work_path()) + 1:]
 
 
@@ -97,7 +94,7 @@ class TableColumnTypes:
 
     def __init__(self, predefined_table_col_types: Dict[str, Dict[str, str]], partition_col_types: Dict[str, str], backend: str):
         self.table_col_types = predefined_table_col_types
-        assert set(partition_col_types.values()).issubset({'int', 'string'}), \
+        assert set(partition_col_types.values()).issubset({'int', 'string', 'Int32', 'String'}), \
             f'only int/string partition column types supported, found {partition_col_types}'
         self.partition_col_types = partition_col_types
         self.backend = backend
@@ -110,20 +107,25 @@ class TableColumnTypes:
             if col_name in self.partition_col_types:
                 return self.partition_col_types[col_name]
             raise Exception(f'column {col_name} not found for table {table_name}, known columns:: {list(self.table_col_types[table_name].keys())}')
+        if self.backend == 'clickhouse':
+            return self.table_col_types[table_name][col_name].strip()
         return self.table_col_types[table_name][col_name].lower().strip()
 
-    def column_types_to_schema(self, backend: Backend, columns: List[str], types: List[str]) -> Union[StructType, List[Col]]:
+    def column_types_to_schema(self, backend: Backend, columns: List[str], types: List[str]) -> Union['StructType', List[Col]]:
         if isinstance(backend, SparkBackend):
             return self.column_types_to_schema_spark(backend.spark, columns, types)
         elif isinstance(backend, RdbBackend):
             return self.column_types_to_schema_rdb(backend, columns, types)
         else:
-            raise Exception(f'unsupported backend type: {type(backend)}')
+            raise AssertionError(f'unsupported backend type: {type(backend)}')
 
     def column_types_to_schema_rdb(self, backend: RdbBackend, columns: List[str], types: List[str]) -> List[Col]:
         return [Col(col, type_) for col, type_ in zip(columns, types)]
 
-    def column_types_to_schema_spark(self, spark: SparkSession, columns: List[str], types: List[str]) -> StructType:
+    def column_types_to_schema_spark(self, spark: 'SparkSession', columns: List[str], types: List[str]) -> 'StructType':
+        from pyspark.sql.functions import expr
+        from pyspark.sql.types import StructType, StructField, IntegerType, ShortType, LongType, DoubleType, FloatType, StringType, BooleanType, \
+            DateType, TimestampType, ArrayType
         fields_creators = {
             'int': lambda name: StructField(name, IntegerType()),
             'tinyint': lambda name: StructField(name, ShortType()),
@@ -161,7 +163,14 @@ class TableColumnTypes:
 
     def cast_as_type(self, table_name: str, col_name: str, col_value: Any, date_converter: Callable = None, col_type: str = None) -> Tuple[str, Any]:
         col_type = self.get_col_type(table_name, col_name) if col_type is None else col_type
-
+        if self.backend == 'clickhouse':
+            if 'Nested' in col_type:
+                raise AssertionError("clickhouse backend can not support Nested col type")
+            # for these type in clickhouse is case insensitive
+            if col_type.lower() in ['bool', 'date', 'datetime', 'decimal']:
+                col_type = col_type.lower()
+        else:
+            col_type = col_type.lower()
         if col_value is None or (isinstance(col_value, str) and col_value.strip() == 'null'):
             return col_type, None
 
@@ -181,7 +190,7 @@ class TableColumnTypes:
                 else:
                     try:
                         # try convert partition column to int to remove possible float values
-                        return col_type, str(int(col_value)) if col_type == 'string' else int(col_value)
+                        return col_type, str(round(float(col_value))) if col_type == 'string' else round(float(col_value))
                     except ValueError as e:
                         if col_type == 'string':
                             return col_type, str(col_value)
@@ -189,31 +198,31 @@ class TableColumnTypes:
                             raise e
 
             if col_type.replace(' ', '').startswith('map<'):
-                raise Exception(f'map type not supported right now when parsing value `{col_value}` for column: {table_name}.{col_name}')
+                raise AssertionError(f'map type not supported right now when parsing value `{col_value}` for column: {table_name}.{col_name}')
             if col_type.startswith('decimal(') or col_type == 'double':
                 return col_type, float(col_value)
-            if col_type in ['bigint', 'int', 'tinyint']:
+            if col_type in ['bigint', 'int', 'tinyint', 'Int32', 'Int64', 'UInt32', 'UInt64']:
                 return col_type, int(col_value)
-            if col_type in ['boolean']:
+            if col_type in ['boolean', 'bool']:
                 if str(col_value).lower() == 'true':
                     return col_type, True
                 elif str(col_value).lower() == 'false':
                     return col_type, False
                 return col_type, bool(col_value)
-            if col_type in ['string', 'text']:
+            if col_type in ['string', 'text', 'String']:
                 return col_type, str(col_value).strip()
-            if col_type.replace(' ', '') in ['array<string>', 'text[]']:
+            if col_type.replace(' ', '') in ['array<string>', 'text[]', 'Array(String)']:
                 return col_type, [s.strip() for s in str(col_value).strip().split('|') if s.strip()]
-            if col_type.replace(' ', '') in ['array<int>', 'int[]']:
+            if col_type.replace(' ', '') in ['array<int>', 'int[]', 'Array(Int)']:
                 return col_type, [int(s.strip()) for s in str(col_value).strip().split('|') if s.strip()]
             if col_type.replace(' ', '') == 'struct<latest_value:string,first_show_time:timestamp>':
                 vals = str(col_value).strip().split('|')
                 if len(vals) < 2:
-                    raise Exception(f'must provide all the values of type {col_type} for {table_name}.{col_name}. Incomplete value `{col_value}`')
+                    raise AssertionError(f'must provide all the values of type {col_type} for {table_name}.{col_name}. Incomplete value `{col_value}`')
                 latest_value = str(vals[0]).strip()
                 first_show_time = str(vals[1]).strip() if not date_converter else date_converter(vals[1])
                 return col_type, (latest_value, first_show_time)
-            if col_type in ['date', 'timestamp']:
+            if col_type in ['date', 'timestamp', 'datetime']:
                 if not date_converter:
                     return col_type, str(col_value).strip()
                 else:
@@ -221,11 +230,11 @@ class TableColumnTypes:
         except Exception as e:
             raise Exception(f'convert {col_type} failed for {table_name}.{col_name} for value `{col_value}`', e)
 
-        raise Exception(f'not supported type {col_type} for column {table_name}.{col_name} when parsing value {col_value}')
+        raise AssertionError(f'not supported type {col_type} for column {table_name}.{col_name} when parsing value {col_value}')
 
 
 class TestCase:
-    def __init__(self, sql_file_path: str = None, sql_file_content: str = None):
+    def __init__(self, sql_file_path: str = None, sql_file_content: str = None, default_col_type='string'):
         self.name, self.vars = None, {}
         self.includes = {}
         self.inputs: List[TableData] = []
@@ -235,6 +244,7 @@ class TestCase:
         self.sql_file_content = sql_file_content
         self.udf_file_paths: List[str] = []
         self.func_file_paths: List[str] = []
+        self.default_col_type = default_col_type
 
     def as_dict(self):
         data = dict([(attr, getattr(self, attr)) for attr in dir(self) if not attr.startswith('_') and not callable(getattr(self, attr))])
@@ -277,14 +287,14 @@ class TestCase:
         elif label == 'FUNCS':
             self.parse_funcs(row_start_idx, rows)
         else:
-            raise Exception(f'unknown label: {label}')
+            raise AssertionError(f'unknown label: {label}')
 
     def parse_includes(self, wb: Book, row_start_idx: int, rows: List[List[Cell]]):
         for i, row in enumerate(rows):
             include_name, include_value = row[1].value.strip(), row[2].value.strip()
             if include_name:
                 if not include_value:
-                    raise Exception(f'parse test case at row {row_start_idx + 1 + i} failed: there must be value set for INCLUDES, found None')
+                    raise AssertionError(f'parse test case at row {row_start_idx + 1 + i} failed: there must be value set for INCLUDES, found None')
                 self.includes[include_name] = include_value
                 log_debug(f'find includes at row {row_start_idx + 1}: include_name={include_name}, include_value={include_value}')
 
@@ -302,7 +312,7 @@ class TestCase:
 
     def parse_vars(self, wb: Book, row_start_idx: int, rows: List[List[Cell]]):
         if len(rows) < 2:
-            raise Exception(f'parse test case at row {row_start_idx + 1} failed: there must be value set for VARS, found None')
+            raise AssertionError(f'parse test case at row {row_start_idx + 1} failed: there must be value set for VARS, found None')
         else:
             for var_name, var_value in zip(rows[0][1:], rows[1][1:]):
                 if var_name.value and var_name.value.strip():
@@ -324,7 +334,7 @@ class TestCase:
         elif isinstance(value, str) and value.strip() != '':
             value = value.strip()
             if len(value) != len('2000-01-01') and len(value) != len('2000-01-01 00:00:00'):
-                raise Exception('date column must be of format `yyyy-MM-dd` or `yyyy-MM-dd HH:mm:ss`')
+                raise AssertionError('date column must be of format `yyyy-MM-dd` or `yyyy-MM-dd HH:mm:ss`')
             format = '%Y-%m-%d' if len(value) == len('2000-01-01') else '%Y-%m-%d %H:%M:%S'
             return datetime.strptime(value, format)
         return datetime(*xlrd.xldate_as_tuple(value, wb.datemode))
@@ -338,9 +348,9 @@ class TestCase:
     def parse_table(self, wb: Book, label: str, row_start_idx: int, rows: List[List[Cell]], table_column_types: TableColumnTypes) -> TableData:
         cells = rows[0]
         if len(cells) < 2 or not cells[1].value or not str(cells[1].value).strip():
-            raise Exception(f'parse test case at row {row_start_idx + 1} failed: there must be table name set for {label}, found None')
+            raise AssertionError(f'parse test case at row {row_start_idx + 1} failed: there must be table name set for {label}, found None')
         if len(cells) < 3 or not cells[2].value or not str(cells[2].value).strip():
-            raise Exception(f'parse test case at row {row_start_idx + 1} failed: there must be at least one column set for {label}, found 0')
+            raise AssertionError(f'parse test case at row {row_start_idx + 1} failed: there must be at least one column set for {label}, found 0')
 
         table_name = str(cells[1].value)
 
@@ -357,7 +367,7 @@ class TestCase:
                     column_types.append(column_name[column_name.index(':') + 1:])
                 else:
                     columns.append(column_name)
-                    column_types.append('string')  # if no type specified, use string as default
+                    column_types.append(self.default_col_type)  # if no type specified, use string as default
             else:
                 columns.append(column_name)
                 if '.' in table_name:
@@ -369,16 +379,18 @@ class TestCase:
             has_values = any([value_cells[i].value not in [None, ''] for i in range(len(columns))])
             if cells[1].value and str(cells[1].value).strip():
                 value_descriptions.append(str(cells[1].value).strip())
-                values.append(self._parse_table_row_values(wb, table_name, columns, row_idx, row_start_idx, value_cells, column_types, table_column_types))
+                values.append(
+                    self._parse_table_row_values(wb, table_name, columns, row_idx, row_start_idx, value_cells, column_types, table_column_types))
             elif label == 'INPUT':
                 # If no description mentioned for this row, just ignore it.
                 # This ensures we must add description for a row, to clarify how the data comes.
                 if has_values:
-                    raise Exception(f'no description for table({table_name}) data at row {row_start_idx + 1 + row_idx + 1}')
+                    raise AssertionError(f'no description for table({table_name}) data at row {row_start_idx + 1 + row_idx + 1}')
             elif label == 'OUTPUT':
                 has_values = any([value_cells[i].value not in [None, ''] for i in range(len(columns))])
                 if has_values:
-                    values.append(self._parse_table_row_values(wb, table_name, columns, row_idx, row_start_idx, value_cells, column_types, table_column_types))
+                    values.append(
+                        self._parse_table_row_values(wb, table_name, columns, row_idx, row_start_idx, value_cells, column_types, table_column_types))
 
         log_debug(f'find {label.lower()} table at row {row_start_idx + 1}: {table_name}{{{", ".join(columns)}}}, data length is {len(values)}')
         print(f'table data for `{table_name}`: ')
@@ -416,7 +428,7 @@ class TestCase:
             return self.sql_file_content
         else:
             if self.sql_file_path is None:
-                raise Exception(f'can not find the sql file having same name with test file')
+                raise AssertionError(f'can not find the sql file having same name with test file')
             with open(work_path.path(self.sql_file_path), 'r') as f:
                 return f.read()
 
@@ -438,10 +450,11 @@ class TestCase:
 
 class TestDataFile:
 
-    def __init__(self, test_data_file: str, sql_reader: 'SqlReader'):
+    def __init__(self, test_data_file: str, sql_reader: 'SqlReader', backend: str='spark'):
         self.test_data_file = test_data_file
         self.sql_reader = sql_reader
         self.wb = open_workbook(self.test_data_file)
+        self.backend = backend
 
     def parse_test_cases(self, table_column_types: TableColumnTypes) -> List[TestCase]:
         wb = self.wb
@@ -473,8 +486,12 @@ class TestDataFile:
         return cases
 
     def parse_test_case(self, case_start_idx: int, case_rows: List[List[Cell]], table_column_types: TableColumnTypes) -> TestCase:
+        if self.backend == 'clickhouse':
+            default_col_type = 'String'
+        else:
+            default_col_type = 'string'
         if self.sql_reader.read_as_content(self.test_data_file):
-            case = TestCase(sql_file_content=self.sql_reader.read_sql(self.test_data_file))
+            case = TestCase(sql_file_content=self.sql_reader.read_sql(self.test_data_file), default_col_type=default_col_type)
         else:
             case = TestCase(self.sql_reader.find_file_path(self.test_data_file[:self.test_data_file.rindex('.')] + '.sql'))
         last_label, last_label_idx = None, -1
@@ -483,13 +500,14 @@ class TestDataFile:
             label = cells[0].value and cells[0].value.strip()
             if label in ['CASE', 'VARS', 'INCLUDES', 'INPUT', 'OUTPUT', 'UDFS', 'FUNCS']:
                 if last_label is not None:
-                    case.parse_test_case_of_label(self.wb, last_label, case_start_idx + last_label_idx, case_rows[last_label_idx:i], table_column_types)
+                    case.parse_test_case_of_label(self.wb, last_label, case_start_idx + last_label_idx, case_rows[last_label_idx:i],
+                                                  table_column_types)
                 last_label, last_label_idx = label, i
         if last_label:
             case.parse_test_case_of_label(self.wb, last_label, case_start_idx + last_label_idx, case_rows[last_label_idx:], table_column_types)
 
         if not case.completed:
-            raise Exception(f'parse test case failed, got incomplete case, missed fields: {case.missed_fields}')
+            raise AssertionError(f'parse test case failed, got incomplete case, missed fields: {case.missed_fields}')
 
         return case
 
@@ -557,13 +575,17 @@ class TestCaseRunner:
         sql = case.read_sql_content()
 
         backend = self.backend_creator(case)
-        self.create_inputs(case, backend)
+        try:
+            self.clean(case, backend)
+            self.create_inputs(case, backend)
 
-        sql_processor = self.create_sql_processor(backend, case, sql)
+            sql_processor = self.create_sql_processor(backend, case, sql)
 
-        sql_processor.run(self.dry_run)
+            sql_processor.run(self.dry_run)
 
-        self.verify_outputs(backend, case)
+            self.verify_outputs(backend, case)
+        finally:
+            backend.clean()
 
     def verify_outputs(self, backend: Backend, case: TestCase):
         tempviews = backend.temp_tables()
@@ -583,7 +605,7 @@ class TestCaseRunner:
             print('actual output: ', list_item_to_set(actual_output))
             self.unit_test_case.assertListEqual(list_item_to_set(expected_output), list_item_to_set(actual_output))
 
-    def create_sql_processor(self, backend: Backend, case: TestCase, sql: str):
+    def create_sql_processor(self, backend: Backend, case: TestCase, sql: str) -> SqlProcessor:
         sql_processor = self.sql_processor_creator(backend, sql, case)
         for udf_file in case.udf_file_paths:
             sql_processor.register_udfs_from_pyfile(work_path.path(udf_file))
@@ -614,7 +636,8 @@ class TestCaseRunner:
         else:
             tempview_name = [tv for tv in tempviews if tv == output.name]
         if len(tempview_name) == 0:
-            raise Exception(f'output `{output.name}` not found after execute test: {case.simple_sql_name}.{case.name}')
+            raise Exception(f'output `{output.name}` not found after execute test: {case.simple_sql_name}.{case.name}. '
+                            f'All temporary views are: {tempviews}')
         elif len(tempview_name) > 1:
             raise Exception(f'multiple temp views found for output `{output.name}` found after execute test: {tempview_name}')
         else:
@@ -632,11 +655,30 @@ class TestCaseRunner:
                 print(f'creating temp table: {input.name}')
                 backend.create_temp_table_with_data(input.name, input.values, schema)
 
+    def clean(self, case: TestCase, backend: Backend):
+        databases = set()
+        table_names = set.union(set(input.name for input in case.inputs), set(output.name for output in case.outputs))
+        for table_name in table_names:
+            if '.' in table_name:
+                databases.add(table_name.split(".")[0])
+        for db in databases:
+            try:
+                if backend.is_bigquery_backend:
+                    backend.exec_native_sql(f"drop schema if exists {db} cascade")
+                elif backend.is_clickhouse_backend:
+                    backend.exec_native_sql(f"drop database if exists {db}")
+            except Exception as e:
+                # BigQuery will throw an exception when deleting a nonexistent dataset even if using [IF EXISTS]
+                import re
+                if not re.match(r"[\s\S]*Permission bigquery.datasets.delete denied on dataset[\s\S]*(or it may not exist)[\s\S]*",
+                                str(e.args[0])):
+                    raise e
+
 
 class SqlTester:
 
     def __init__(self,
-                 backend_creator: Callable[[TestCase], Backend],
+                 backend_creator: Callable[[TestCase], Backend] = None,
                  table_column_types: TableColumnTypes = None,
                  sql_reader_creator: Callable[[], 'SqlReader'] = None,
                  sql_processor_creator: Callable[[Backend, str, TestCase], SqlProcessor] = None,
@@ -653,12 +695,13 @@ class SqlTester:
             return SqlProcessor(backend, sql, [], variables=vars, scala_udf_initializer=scala_udf_initializer, includes=case.includes)
 
         self.sql_processor_creator = sql_processor_creator or create_sql_processor
+        self.backend = backend
         self.table_column_types = table_column_types or TableColumnTypes({}, {}, backend)
         self.unit_test_case = unit_test_case
         self.dry_run = dry_run
         self.env = env
         self.sql_reader = sql_reader_creator() if sql_reader_creator else SqlReader()
-        self.test_case_runner = TestCaseRunner(self.env, self.dry_run, backend_creator, table_column_types,
+        self.test_case_runner = TestCaseRunner(self.env, self.dry_run, backend_creator, self.table_column_types,
                                                sql_processor_creator=self.sql_processor_creator, unit_test_case=self.unit_test_case)
 
     def run_tests(self, test_data_files: List[str]):
@@ -677,7 +720,7 @@ class SqlTester:
         cases = self.parse_test_cases(test_data_file, self.table_column_types)
 
         if case_idx != -1 and (case_idx < 0 or case_idx >= len(cases)):
-            raise Exception(f'test case {case_idx} not found. {len(cases)} cases in {test_data_file} are found.')
+            raise AssertionError(f'test case {case_idx} not found. {len(cases)} cases in {test_data_file} are found.')
         cases = cases if case_idx == -1 else [cases[case_idx]]
 
         tr = TestResult(test_data_file)
@@ -688,13 +731,13 @@ class SqlTester:
 
     def parse_test_cases(self, test_data_file, table_column_types: TableColumnTypes):
         if test_data_file.endswith('.xlsx'):
-            cases = TestDataFile(test_data_file, sql_reader=self.sql_reader).parse_test_cases(table_column_types)
+            cases = TestDataFile(test_data_file, sql_reader=self.sql_reader, backend=self.backend).parse_test_cases(table_column_types)
         elif test_data_file.endswith('.json'):
             with open(test_data_file, 'r') as f:
                 cases = json.loads(f.read(), object_hook=json_util.object_hook)
             cases = [TestCase.from_dict(case_dict) for case_dict in cases]
         else:
-            raise Exception(f'unsupported format of test file: {test_data_file}')
+            raise AssertionError(f'unsupported format of test file: {test_data_file}')
         return cases
 
     def run_case(self, case: TestCase) -> bool:
@@ -708,7 +751,7 @@ class SqlTester:
 
     def convert_cases_to_json(self, test_data_file: str):
         if not test_data_file.endswith('.xlsx'):
-            raise Exception(f'only support to convert excel file, got `{test_data_file}`')
+            raise AssertionError(f'only support to convert excel file, got `{test_data_file}`')
         cases = TestDataFile(test_data_file, sql_reader=self.sql_reader).parse_test_cases(self.table_column_types)
         data = [case.as_dict() for case in cases]
         output_file = test_data_file[:-len('.xlsx')] + '.json'
@@ -729,7 +772,7 @@ from {self.__module__} import SqlTester
 
 
 class SqlTest(unittest.TestCase):
-    
+
     def __init__(self, methodName='runTest'):
         super().__init__(methodName)
         work_dir = os.environ.get('WORK_DIR')
@@ -742,7 +785,7 @@ class SqlTest(unittest.TestCase):
     def test_{{{{loop.index}}}}(self):
         # {{{{ case.name }}}}
         self.assertTrue(self.sql_tester.run_test(self.test_data_file, {{{{loop.index0}}}}).is_success)
-        
+
 {{% endfor %}}
 if __name__ == '__main__':
     unittest.main()
@@ -769,4 +812,3 @@ class SqlReader:
 
     def read_as_file_path(self, test_data_file: str):
         return not self.read_as_content(test_data_file)
-
